import transaction

from onegov.winterthur.collections import AddressCollection
from onegov_testing import Client as BaseClient


class Client(BaseClient):
    skip_first_form = True


def test_view_addresses(winterthur_app, streets_csv, addresses_csv):
    client = Client(winterthur_app)

    page = client.get('/streets')
    assert "Keine Strassen gefunden" in page

    transaction.begin()

    addresses = AddressCollection(winterthur_app.session())
    addresses.import_from_csv(streets_csv, addresses_csv)

    transaction.commit()

    page = client.get('/streets')
    assert "Zürcherstrasse" in page
    page = page.click("Zürcherstrasse")

    assert "Zürcherstrasse 100" in page
    assert "Schlosstal" in page
    assert "Töss" in page



import textwrap

from .utils import as_choices
from .utils import build_metadata
from .utils import load_files_by_prefix
from .utils import store_in_zip
from .utils import Geocoder


def transform_fachstellen(path, prefix, output_file):
    data = []

    for r in load_files(path, prefix).lines:
        data.append({
            'Fachstelle/Name': r.name,
            'Fachstelle/Beschreibung': r.description,
            'Fachstelle/Adresse': '\n'.join(
                l.strip() for l in r.company.split('</br>') if l.strip()
            ),
            'Fachstelle/Telefon': r.phone,
            'Fachstelle/E-Mail': r.email,
            'Fachstelle/Webseite': r.www and f'http://{r.www}' or '',
            'Fachstelle/Themen': [
                g.strip() for g in r.addressgroup.split(';')
            ]
        })

    topics = {g for d in data for g in d['Fachstelle/Themen']}
    topics = as_choices(sorted(list(topics)))

    geo = Geocoder()

    for entry in data:
        location = geo.geocode(
            '\n'.join(entry['Fachstelle/Adresse'].split('\n')[-2:]))

        entry['Latitude'] = location.latitude
        entry['Longitude'] = location.longitude

    metadata = build_metadata(
        title='Fachstellen A-Z',
        lead='Die Fachstellen der Stadt Winterthur',
        title_format='[Fachstelle/Name]',
        lead_format='[Fachstelle/Beschreibung]',
        content_fields=['Fachstelle/Themen'],
        contact_fields=[
            'Fachstelle/Adresse',
            'Fachstelle/Telefon',
            'Fachstelle/E-Mail',
            'Fachstelle/Webseite',
        ],
        keyword_fields=[
            'Fachstelle/Themen'
        ],
        structure=textwrap.dedent(f"""\
            # Fachstelle
            Name *= ___
            Beschreibung = ...

            Adresse = ...
            Telefon = ___
            E-Mail = @@@
            Webseite = ___

            Themen =
                {topics}
        """))

    store_in_zip(output_file, {
        'metadata.json': metadata,
        'data.json': data
    })


def load_files(path, prefix):
    return load_files_by_prefix(path, prefix, mapping=(
        ('address_fachstellen', (
            'last_name',
            'name',
            'phone',
            'www',
            'email',
            'company',
            'city',
            'zip',
            'description',
            'addressgroup'
        )),
    ))



import math
import pytest

from onegov.winterthur.collections import AddressCollection


@pytest.mark.xfail(reason="the remote host providing the csv might be down")
def test_download_addresses(session):
    addresses = AddressCollection(session)
    addresses.update()

    assert addresses.query().count() >= 18_000


def test_update_addresses(session, streets_csv, addresses_csv):
    addresses = AddressCollection(session)
    addresses.import_from_csv(streets_csv, addresses_csv)

    count = addresses.query().count()

    # check a rough address count
    assert count >= 18_000

    # check a well known record
    a = addresses.query().first()
    assert a.id == 9236
    assert a.street == 'Ackeretstrasse'
    assert a.house_number == 1
    assert a.house_extra == ''
    assert a.zipcode == 8400
    assert a.zipcode_extra == 0
    assert a.place == 'Winterthur'
    assert a.district == 'Winterthur-Stadt'

    # synchronise again -> the count should roughly stay the same
    # if not, the file on the remote end has been updated!
    addresses.update()
    assert math.isclose(count, addresses.query().count(), rel_tol=0.1)

    # check the encoding
    a = addresses.query().filter_by(street="Alte Römerstrasse").first()
    assert a.street == "Alte Römerstrasse"



import pytest
import transaction

from onegov.core.csv import CSVFile
from onegov.core.utils import module_path
from onegov.user import User
from onegov.winterthur import WinterthurApp
from onegov.winterthur.initial_content import create_new_organisation
from onegov_testing.utils import create_app
from pathlib import Path


@pytest.fixture()
def fixtures():
    return Path(module_path('onegov.winterthur', 'tests/fixtures'))


@pytest.fixture(scope='session', autouse=True)
def scan_dependencies():
    import importscan
    import onegov

    importscan.scan(onegov.org, ignore=['.test', '.tests'])


@pytest.fixture()
def streets_csv(fixtures):
    with (fixtures / 'streets.csv').open('rb') as f:
        yield CSVFile(f)


@pytest.fixture()
def addresses_csv(fixtures):
    with (fixtures / 'addresses.csv').open('rb') as f:
        yield CSVFile(f)


@pytest.yield_fixture(scope='function')
def winterthur_app(request):
    yield create_winterthur_app(request, use_elasticsearch=False)


@pytest.yield_fixture(scope='function')
def es_winterthur_app(request):
    yield create_winterthur_app(request, use_elasticsearch=True)


def create_winterthur_app(request, use_elasticsearch):

    app = create_app(
        app_class=WinterthurApp,
        request=request,
        use_elasticsearch=use_elasticsearch)

    session = app.session()

    org = create_new_organisation(app, name="Winterthur")
    org.meta['reply_to'] = 'mails@govikon.ch'
    org.meta['locales'] = 'de_CH'

    # usually we don't want to create the users directly, anywhere else you
    # *need* to go through the UserCollection. Here however, we can improve
    # the test speed by not hashing the password for every test.
    test_password = request.getfixturevalue('test_password')

    session.add(User(
        username='admin@example.org',
        password_hash=test_password,
        role='admin'
    ))
    session.add(User(
        username='editor@example.org',
        password_hash=test_password,
        role='editor'
    ))

    transaction.commit()
    session.close_all()

    return app



from onegov.winterthur.app import WinterthurApp
from onegov.winterthur.collections import AddressCollection
from onegov.winterthur.collections import AddressSubsetCollection


@WinterthurApp.path(
    model=AddressCollection,
    path='/streets')
def get_streets_directory(app):
    return AddressCollection(app.session())


@WinterthurApp.path(
    model=AddressSubsetCollection,
    path='/streets/{street}')
def get_street_subset(app, street):
    subset = AddressSubsetCollection(app.session(), street=street)
    return subset.exists() and subset or None



from cached_property import cached_property
from onegov.org.layout import DefaultLayout
from onegov.org.new_elements import Link, Intercooler
from onegov.winterthur import _
from onegov.winterthur.collections import AddressCollection


class AddressLayout(DefaultLayout):

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Addresses"), '#'),
        ]

    @cached_property
    def editbar_links(self):
        if not self.request.is_manager:
            return

        return [
            Link(
                text=_("Update"),
                url=self.csrf_protected_url(
                    self.request.link(self.model, '+update')
                ),
                attrs={'class': 'sync'},
                traits=Intercooler(
                    request_method='POST',
                    redirect_after=self.request.url
                )
            )
        ]


class AddressSubsetLayout(DefaultLayout):

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Addresses"), self.request.class_link(AddressCollection)),
            Link(_(self.model.street), '#')
        ]



from onegov.core.orm import Base
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import Text


class WinterthurAddress(Base):

    __tablename__ = 'winterthur_addresses'

    #: the adress id
    id = Column(Integer, nullable=False, primary_key=True)

    #: the street
    street_id = Column(Integer, nullable=False)
    street = Column(Text, nullable=False)

    #: the house
    house_number = Column(Integer, nullable=False)
    house_extra = Column(Text, nullable=True)

    #: the place
    zipcode = Column(Integer, nullable=False)
    zipcode_extra = Column(Integer, nullable=True)

    place = Column(Text, nullable=False)

    #: the district
    district = Column(Text, nullable=False)

    #: the neighbourhood
    neighbourhood = Column(Text, nullable=False)

    @classmethod
    def as_addressless(cls, street_id, street):
        return cls(
            street_id=street_id,
            street=street,
            house_number=-1,
            zipcode=-1,
            place="",
            district="",
            neighbourhood="",
        )

    @property
    def is_addressless(self):
        return self.house_number == -1

    @property
    def title(self):
        if self.is_addressless:
            return self.street
        else:
            return f"{self.street} {self.house_number}{self.house_extra}"



import re

from onegov.core import utils
from onegov.org import OrgApp
from onegov.org.app import get_common_asset as get_org_common_asset
from onegov.org.app import get_i18n_localedirs as get_org_i18n_localedirs
from onegov.winterthur.initial_content import create_new_organisation
from onegov.winterthur.theme import WinterthurTheme


class WinterthurApp(OrgApp):

    #: the version of this application (do not change manually!)
    version = '0.2.11'

    frame_ancestors = {
        'https://winterthur.ch',
        'https://*.winterthur.ch'
    }

    def configure_organisation(self, **cfg):
        cfg.setdefault('enable_user_registration', False)
        cfg.setdefault('enable_yubikey', False)
        super().configure_organisation(**cfg)

    def enable_iframes(self, request):
        request.content_security_policy.frame_ancestors |= self.frame_ancestors
        request.include('iframe-resizer')


@WinterthurApp.tween_factory()
def enable_iframes_tween_factory(app, handler):
    iframe_paths = (
        r'/streets.*',
        r'/director(y|ies|y-submission/.*)',
        r'/ticket/.*'
    )

    iframe_paths = re.compile(rf"({'|'.join(iframe_paths)})")

    def enable_iframes_tween(request):
        """ Enables iframes on matching paths. """

        result = handler(request)

        if iframe_paths.match(request.path_info):
            request.app.enable_iframes(request)

        return result

    return enable_iframes_tween


@WinterthurApp.template_directory()
def get_template_directory():
    return 'templates'


@WinterthurApp.setting(section='core', name='theme')
def get_theme():
    return WinterthurTheme()


@WinterthurApp.setting(section='org', name='create_new_organisation')
def get_create_new_organisation_factory():
    return create_new_organisation


@WinterthurApp.setting(section='org', name='default_directory_search_widget')
def get_default_directory_search_widget():
    return 'inline'


@WinterthurApp.setting(section='i18n', name='localedirs')
def get_i18n_localedirs():
    mine = utils.module_path('onegov.winterthur', 'locale')
    return [mine] + get_org_i18n_localedirs()


@WinterthurApp.webasset_path()
def get_js_path():
    return 'assets/js'


@WinterthurApp.webasset_output()
def get_webasset_output():
    return 'assets/bundles'


@WinterthurApp.webasset('street-search')
def get_search_asset():
    yield 'wade.js'
    yield 'string-score.js'
    yield 'street-search.js'


@WinterthurApp.webasset('iframe-resizer')
def get_iframe_resizer():
    yield 'iframe-resizer-options.js'
    yield 'iframe-resizer-contentwindow.js'


@WinterthurApp.webasset('common')
def get_common_asset():
    yield from get_org_common_asset()
    yield 'winterthur.js'



from onegov.winterthur import WinterthurApp
from onegov.org.views import directory as base
from onegov.directory import DirectoryCollection, DirectoryEntryCollection
from onegov.core.security import Secret
from onegov.form import merge_forms
from onegov.form.utils import get_fields_from_class


def get_directory_form_class(model, request):
    forms = [base.get_directory_form_class(model, request)]
    registry = request.app.config.directory_search_widget_registry

    fields = {}

    for name, cls in registry.items():
        if hasattr(cls, 'form'):
            fields[name] = tuple(n for n, f in get_fields_from_class(cls.form))
            forms.append(cls.form)

    class AdaptedDirectoryForm(merge_forms(*forms)):

        def populate_obj(self, obj, *args, **kwargs):
            nonlocal fields

            super().populate_obj(obj, *args, **kwargs)

            obj.searchwidget_config = {}

            for name, fields in fields.items():
                obj.searchwidget_config[name] = {
                    f: self.data[f] for f in fields
                }

        def process_obj(self, obj):
            nonlocal fields

            super().process_obj(obj)

            if not obj.searchwidget_config:
                return

            for name, fields in fields.items():
                if name not in obj.searchwidget_config:
                    continue

                for f in fields:
                    getattr(self, f).data\
                        = obj.searchwidget_config[name].get(f)

    return AdaptedDirectoryForm


@WinterthurApp.form(model=DirectoryCollection, name='new', template='form.pt',
                    permission=Secret, form=get_directory_form_class)
def handle_new_directory(self, request, form):
    return base.handle_new_directory(self, request, form)


@WinterthurApp.form(model=DirectoryEntryCollection, name='edit',
                    template='directory_form.pt', permission=Secret,
                    form=get_directory_form_class)
def handle_edit_directory(self, request, form):
    return base.handle_edit_directory(self, request, form)



from concurrent.futures import ThreadPoolExecutor
from io import BytesIO
from onegov.core.collection import GenericCollection
from onegov.core.csv import CSVFile
from onegov.core.orm import as_selectable
from onegov.winterthur.models import WinterthurAddress
from pycurl import Curl
from sqlalchemy import select, func

HOST = 'https://stadt.winterthur.ch'
STREETS = f'{HOST}/_static/strassenverzeichnis/gswpl_strver_str.csv'
ADDRESSES = f'{HOST}/_static/strassenverzeichnis/gswpl_strver_adr.csv'


class AddressCollection(GenericCollection):

    @property
    def model_class(self):
        return WinterthurAddress

    def streets(self):
        query = as_selectable("""
            SELECT
                UPPER(UNACCENT(LEFT(street, 1))) AS letter, -- Text
                street                                      -- Text
            FROM
                winterthur_addresses
            GROUP BY
                street
            ORDER BY
                unaccent(street)
        """)

        return self.session.execute(select(query.c))

    def update(self, streets=STREETS, addresses=ADDRESSES):
        self.delete_existing()
        self.import_from_csv(*self.load_urls(streets, addresses))

    def delete_existing(self):
        for address in self.query():
            self.session.delete(address)

    def import_from_csv(self, streets, addresses):
        streets = {s.strc: s.bez for s in streets.lines}

        addressless = set(streets.keys())
        max_id = 0

        for r in addresses.lines:
            addressless.discard(r.strc)

            address = WinterthurAddress()
            address.id = int(r.einid)
            address.street_id = int(r.strc)
            address.street = streets[r.strc]
            address.house_number = int(r.hnr)
            address.house_extra = r.hnrzu
            address.zipcode = int(r.plz)
            address.zipcode_extra = r.plzzu
            address.place = r.ort
            address.district = r.kreisname
            address.neighbourhood = r.quartiername

            self.session.add(address)

            max_id = max(max_id, address.id)

        # some streets do not have addresses -> we write a special record for
        # those streets so they still show up in our UI
        #
        # not the most elegant solution, but better than introducing a separate
        # table at least for now
        for id, key in enumerate(addressless, start=max_id + 1):
            address = self.model_class.as_addressless(int(key), streets[key])
            address.id = id

            self.session.add(address)

        self.session.flush()

    def load_urls(self, *urls):
        with ThreadPoolExecutor(max_workers=2) as executor:
            futures = (executor.submit(self.load_url, url) for url in urls)
            return tuple(f.result() for f in futures)

    def load_url(self, url):
        buffer = BytesIO()

        c = Curl()
        c.setopt(c.URL, url)
        c.setopt(c.WRITEDATA, buffer)
        c.perform()
        c.close()

        return CSVFile(buffer)


class AddressSubsetCollection(GenericCollection):

    def __init__(self, session, street):
        super().__init__(session)
        self.street = street

    @property
    def model_class(self):
        return WinterthurAddress

    def subset(self):
        subset = self.query().filter_by(street=self.street)

        return subset.order_by(
            func.unaccent(WinterthurAddress.street),
            WinterthurAddress.house_number,
            WinterthurAddress.house_extra
        )

    def exists(self):
        return self.session.query(self.subset().exists()).scalar()



from cached_property import cached_property
from elasticsearch_dsl.query import MultiMatch
from onegov.core.templates import render_macro
from onegov.directory import DirectoryEntry, DirectoryEntryCollection
from onegov.form import as_internal_id
from onegov.winterthur.app import WinterthurApp
from sqlalchemy import cast, func, literal_column, Text
from sqlalchemy.dialects.postgresql import array


def lines(value):
    if isinstance(value, (tuple, list)):
        yield from value

    for line in str(value).split('\n'):
        yield line


@WinterthurApp.directory_search_widget('inline')
class InlineSearch(object):

    def __init__(self, request, directory, search_query):
        self.app = request.app
        self.request = request
        self.directory = directory
        self.search_query = search_query

    @cached_property
    def term(self):
        return (self.search_query or {}).get('term', None)

    @cached_property
    def searchable(self):
        return tuple(self.directory.configuration.searchable)

    @cached_property
    def hits(self):
        if not self.term:
            return None

        search = self.app.es_search_by_request(
            request=self.request,
            types=('extended_directory_entries', )
        )

        search = search.filter('term', directory_id=str(self.directory.id))

        fields = tuple(
            f for f in DirectoryEntry.es_properties.keys()
            if not f.startswith('es_') and not f == 'directory_id'
        )

        match = MultiMatch(query=self.term, fields=fields, fuzziness=1)
        search = search.query(match)

        for field in fields:
            search = search.highlight(field)

        return {hit.meta.id: hit for hit in search[0:100].execute()}

    def html(self, layout):
        return render_macro(layout.macros['inline_search'], self.request, {
            'term': self.term,
            'directory': self.directory,
            'action': self.request.class_link(
                DirectoryEntryCollection,
                variables={
                    'directory_name': self.directory.name,
                    'search': self.name
                }
            )
        })

    def fragments(self, entry):
        for name in self.searchable:
            key = as_internal_id(name)

            fragments = (
                f'{name}: {line.lstrip(" -")}'
                for line in lines(entry.values[key])
                if self.term in line
            )

            fragments = tuple(f for ix, f in enumerate(fragments) if ix <= 2)

            if fragments:
                yield name, fragments

    def lead(self, layout, entry):
        if not self.term:
            return None

        hit = self.hits[str(entry.id)]

        for key in hit.meta.highlight:
            for fragment in hit.meta.highlight[key]:
                return fragment

    def adapt(self, query):
        if not self.term:
            return query

        ids = tuple(self.hits.keys())

        query = query.filter(DirectoryEntry.id.in_(ids))

        if ids:
            query = query.order_by(False)
            query = query.order_by(
                func.array_position(
                    array(ids),
                    cast(literal_column('id'), Text)
                )
            )

        print(query)

        return query



import csv
import json
import re
import shutil
import tempfile
import textwrap

from collections import namedtuple
from geopy.geocoders import Nominatim
from onegov.core.cache import lru_cache
from onegov.core.csv import CSVFile
from onegov.core.html import html_to_text as core_html_to_text
from onegov.core.utils import normalize_for_url, safe_format_keys
from onegov.form import flatten_fieldsets, parse_formcode
from pathlib import Path


Point = namedtuple('Point', ('latitude', 'longitude'))


class Geocoder(object):

    def __init__(self):
        self.nominatim = Nominatim(user_agent='onegov-cloud-import')

    @lru_cache(maxsize=1024)
    def geocode(self, address):
        result = self.nominatim.geocode(address)
        return Point(
            getattr(result, 'latitude', None),
            getattr(result, 'longitude', None)
        )


def contains_number(text):
    return re.search(r'[0-9]+', text) and True or False


def contains_only_numbers(text):
    return re.match(r'^[0-9\s]+$', text) and True or False


def contains_department(text):
    return 'depart' in text.lower()


def contains_street(text):
    return re.match(r'^[a-zA-Zöäü\s]+ [0-9]+', text) and True or False


def contains_location(text):
    return re.match(r'^[0-9]+ [a-zA-Zöäü\s]+', text) and True or False


def contains_website(text):
    return not contains_email(text) and '.ch' in text or 'www.' in text


def contains_email(text):
    return '@' in text


def pop_matching(lines, predicate):
    for ix, line in enumerate(lines):
        if predicate(line):
            break
    else:
        return None

    return lines.pop(ix)


def build_metadata(title, lead, structure, title_format, lead_format,
                   content_fields, contact_fields, keyword_fields,
                   link_pattern=None, link_title=None, **extra):

    print(structure)

    fieldnames = tuple(
        f.human_id for f in flatten_fieldsets(parse_formcode(structure))
    )

    return {
        'type': 'extended',
        'name': normalize_for_url(title),
        'title': title,
        'lead': lead,
        'structure': structure,
        'configuration': {
            'title': title_format,
            'lead': lead_format,
            'display': {
                'contact': contact_fields,
                'content': content_fields,
            },
            'keywords': keyword_fields,
            'searchable': fieldnames,
            'order': safe_format_keys(title_format),
            'link_pattern': link_pattern,
            'link_title': link_title
        },
        **extra
    }


def load_files_by_prefix(path, prefix, mapping):
    files = []

    dialect = csv.excel

    for name, fields in mapping:
        for p in Path(path).glob(f'{prefix}*'):
            if name in p.name:
                files.append(CSVFile(
                    csvfile=p.open('rb'),
                    expected_headers=fields,
                    dialect=dialect
                ))
                break

    assert all(files)
    assert len(files) == len(mapping)

    return files[0] if len(files) == 1 else files


def as_choices(values, indent=16):
    return textwrap.indent(
        '\n'.join(f'[ ] {v}' for v in values),
        ' ' * 16,
        lambda line: not values[0] == line.split(']')[-1].strip()
    )


def store_in_zip(output_file, data):
    with tempfile.TemporaryDirectory() as directory:

        for filename, data in data.items():
            with (Path(directory) / filename).open('w') as f:
                f.write(json.dumps(data))

        shutil.make_archive(output_file.split('.')[0], 'zip', directory)


def html_to_text(html):
    return core_html_to_text(
        html,
        ul_item_mark='•',
        strong_mark='',
        emphasis_mark=''
    ).replace('\n\n', '\n')



import textwrap

from itertools import groupby
from .utils import as_choices
from .utils import build_metadata
from .utils import html_to_text
from .utils import load_files_by_prefix
from .utils import store_in_zip
from .utils import Geocoder


def transform_rooms(path, prefix, output_file):
    rooms = load_files_by_prefix(path, prefix, mapping=(
        ('Raumverzeichnis', (
            "uid",
            "categories",
            "room_name",
            "room_tags",
            "category",
            "size_max",
            "size_min",
            "zone",
            "room_images",
            "renter_present_time",
            "room_sponsorship",
            "room_ability",
            "room_description",
            "room_infrastructure",
            "room_tos",
            "district",
            "room_address",
            "room_zip",
            "room_city",
            "room_phone",
            "room_email",
            "hirer_name",
            "hirer_address",
            "hirer_zip",
            "hirer_city",
            "hirer_phone",
            "hirer_email"
        )),
    ))

    rooms = sorted((r for r in rooms.lines), key=lambda r: r.uid)
    data = []

    for uid, records in groupby(rooms, key=lambda r: r.uid):
        first, *rest = records

        data.append({
            'Raum/Name': first.room_name,
            'Raum/Untertitel': first.room_tags,
            'Raum/Beschreibung': html_to_text(first.room_description),
            'Raum/Plätze': first.size_max,
            'Raum/Adresse': '\n'.join(p.strip() for p in (
                first.room_address,
                f'{first.room_zip} {first.room_city}'
            ) if p.strip()),
            'Raum/Telefon': first.room_phone,
            'Raum/E-Mail': first.room_email,
            'Raum/Funktionen': html_to_text(first.room_ability),
            'Raum/Infrastruktur': html_to_text(first.room_infrastructure),
            'Raum/Kategorie': [first.category] + [r.category for r in rest],
            'Ort/Zone': [first.zone] if first.zone else [],
            'Ort/Kreis': [first.district] if first.district else [],
            'Vermieter/Name': first.hirer_name,
            'Vermieter/Adresse': '\n'.join(p.strip() for p in (
                first.hirer_address,
                f'{first.hirer_zip} {first.hirer_city}'
            ) if p.strip()),
            'Vermieter/Telefon': first.hirer_phone,
            'Vermieter/E-Mail': first.hirer_email,
            'Vermieter/Präsenzzeit': first.renter_present_time,
            'Details/Sponsor': first.room_sponsorship,
            'Details/Nutzungsbedingungen': html_to_text(first.room_tos)
        })

    zones = sorted(
        set(zone for d in data for zone in d['Ort/Zone']))
    districts = sorted(
        set(district for d in data for district in d['Ort/Kreis']))
    categories = sorted(
        set(cat for d in data for cat in d['Raum/Kategorie']))

    metadata = build_metadata(
        title='Räume zum Mieten',
        lead=(
            'In Winterthur gibt es ein vielfältiges Angebot an Räumen, die '
            'gemietet werden können. Freizeitanlagen, Waldhütten, Räume für '
            'Feste, Sitzungen oder kulturelle Veranstaltungen sowie Kurs- und '
            'Seminarräume.'
        ),
        title_format='[Raum/Name]',
        lead_format='[Raum/Beschreibung]',
        content_fields=[
            'Raum/Untertitel',
            'Raum/Plätze',
            'Raum/Adresse',
            'Raum/Telefon',
            'Raum/E-Mail',
            'Raum/Funktionen',
            'Raum/Infrastruktur',
            'Raum/Kategorie',
            'Ort/Zone',
            'Ort/Kreis',
            'Details/Sponsor',
            'Details/Nutzungsbedingungen'
        ],
        contact_fields=[
            'Vermieter/Name',
            'Vermieter/Adresse',
            'Vermieter/Telefon',
            'Vermieter/E-Mail',
            'Vermieter/Präsenzzeit'
        ],
        keyword_fields=[
            'Raum/Kategorie',
            'Ort/Zone',
            'Ort/Kreis'
        ],
        structure=textwrap.dedent(f"""\
            # Raum
            Name *= ___
            Untertitel = ___
            Beschreibung = ...
            Plätze = ___

            Adresse = ...
            Telefon = ___
            E-Mail = ___

            Funktionen = ...
            Infrastruktur = ...

            Kategorie =
                {as_choices(categories)}

            # Ort
            Zone =
                {as_choices(zones)}

            Kreis =
                {as_choices(districts)}

            # Vermieter
            Name = ___
            Adresse = ...
            Telefon = ___
            E-Mail = ___
            Präsenzzeit = ___

            # Details
            Sponsor = ___
            Nutzungsbedingungen = ...

        """))

    geo = Geocoder()

    for entry in data:
        location = geo.geocode(entry['Raum/Adresse'])

        entry['Latitude'] = location.latitude
        entry['Longitude'] = location.longitude

    store_in_zip(output_file, {
        'metadata.json': metadata,
        'data.json': data
    })



import textwrap

from .utils import as_choices
from .utils import build_metadata
from .utils import load_files_by_prefix
from .utils import store_in_zip
from .utils import html_to_text


def transform_theater(path, prefix, output_file):
    columns = (
        'type',
        'category',
        'uid',
        'titel',
        'organizer',
        'beschreibung',
        'geeignet',
        'dauer',
        'datum',
        'kosten',
        'leitung',
        'besonderes',
        'anmeldung',
        'onlineanmeldung',
    )

    records = load_files_by_prefix(path, prefix, (('Theater', columns),))
    data = []

    for record in records.lines:
        data.append({
            'Angebot/Titel': record.titel,
            'Angebot/Beschreibung': html_to_text(record.beschreibung),
            'Angebot/Dauer': record.dauer,
            'Angebot/Kategorie': [record.category],
            'Angebot/Zielgruppe': record.geeignet and [
                record.geeignet.replace(
                    '1.-  5. Klasse',
                    '1. - 5. Klasse'
                ).replace(
                    '6. -  9. Klasse',
                    '6. - 9. Klasse'
                )
            ] or [],
            'Details/Datum': record.datum,
            'Details/Anmeldung': record.anmeldung,
            'Details/Kosten': record.kosten,
            'Details/Leitung': record.leitung,
            'Details/Besonderes': record.besonderes
        })

    categories = sorted(set(
        d['Angebot/Kategorie'][0]
        for d in data if d['Angebot/Kategorie']
    ))
    targets = sorted(set(
        d['Angebot/Zielgruppe'][0]
        for d in data if d['Angebot/Zielgruppe']
    ))

    metadata = build_metadata(
        title="Angebote Theaterpädagogik",
        lead=(
            "Allen Schulklassen werden faszinierende, an- und aufregende, "
            "lehrreiche und berührende Erlebnisse in den Museen, Theatern "
            "und Konzertsälen dieser Stadt angeboten. "
        ),
        title_format='[Angebot/Titel]',
        lead_format='[Angebot/Beschreibung]',
        content_fields=[
            'Angebot/Dauer',
            'Angebot/Kategorie',
            'Angebot/Zielgruppe',
            'Details/Datum',
            'Details/Anmeldung',
            'Details/Kosten',
            'Details/Leitung',
            'Details/Besonderes',
        ],
        contact_fields=[],
        keyword_fields=[
            'Angebot/Kategorie',
            'Angebot/Zielgruppe'
        ],
        structure=textwrap.dedent(f"""\
            # Angebot
            Titel *= ___
            Beschreibung *= ...
            Dauer = ___

            Kategorie *=
                {as_choices(categories)}

            Zielgruppe =
                {as_choices(targets)}

            # Details
            Datum = ___
            Anmeldung = ___
            Kosten = ___
            Leitung = ___
            Besonderes = ...
        """)
    )

    store_in_zip(output_file, {
        'metadata.json': metadata,
        'data.json': data
    })



from onegov.core.utils import module_path
from onegov.org.initial_content import load_content, add_pages
from onegov.org.models import Organisation


def create_new_organisation(app, name, locale='de_CH'):
    assert locale == 'de_CH'

    path = module_path('onegov.winterthur', 'content/de.yaml')
    content = load_content(path)

    org = Organisation(name=name, **content['organisation'])
    org.meta['locales'] = locale

    session = app.session()
    session.add(org)

    add_pages(session, path)

    return org



import click

from .fachstellen import transform_fachstellen
from .museum import transform_museum
from .rooms import transform_rooms
from .theater import transform_theater
from .vollzug import transform_vollzug


POST_MORTEM = False


@click.group()
@click.option('--pdb', default=False, is_flag=True)
def cli(pdb):
    global POST_MORTEM
    POST_MORTEM = pdb


@cli.command(name='transform-vollzug')
@click.argument('path', type=click.Path(exists=True, file_okay=False))
@click.argument('prefix')
@click.option('--output', type=click.Path(), default=None)
def transform_vollzug_cli(path, prefix, output):
    try:
        return transform_vollzug(path, prefix, output)
    except Exception:
        if POST_MORTEM:
            import pdb; pdb.post_mortem()  # noqa
        raise


@cli.command(name='transform-fachstellen')
@click.argument('path', type=click.Path(exists=True, file_okay=False))
@click.argument('prefix')
@click.option('--output', type=click.Path(), default=None)
def transform_fachstellen_cli(path, prefix, output):
    try:
        return transform_fachstellen(path, prefix, output)
    except Exception:
        if POST_MORTEM:
            import pdb; pdb.post_mortem()  # noqa
        raise


@cli.command(name='transform-museum')
@click.argument('path', type=click.Path(exists=True, file_okay=False))
@click.argument('prefix')
@click.option('--output', type=click.Path(), default=None)
def transform_museum_cli(path, prefix, output):
    try:
        return transform_museum(path, prefix, output)
    except Exception:
        if POST_MORTEM:
            import pdb; pdb.post_mortem()  # noqa
        raise


@cli.command(name='transform-rooms')
@click.argument('path', type=click.Path(exists=True, file_okay=False))
@click.argument('prefix')
@click.option('--output', type=click.Path(), default=None)
def transform_rooms_cli(path, prefix, output):
    try:
        return transform_rooms(path, prefix, output)
    except Exception:
        if POST_MORTEM:
            import pdb; pdb.post_mortem()  # noqa
        raise


@cli.command(name='transform-theater')
@click.argument('path', type=click.Path(exists=True, file_okay=False))
@click.argument('prefix')
@click.option('--output', type=click.Path(), default=None)
def transform_theater_cli(path, prefix, output):
    try:
        return transform_theater(path, prefix, output)
    except Exception:
        if POST_MORTEM:
            import pdb; pdb.post_mortem()  # noqa
        raise


# run cli handler when running this thorugh pyhton -m
cli()



from onegov.winterthur.app import WinterthurApp
from onegov.winterthur.collections import AddressCollection


@WinterthurApp.cronjob(hour=15, minute=50, timezone='Europe/Zurich')
def update_streets_directory(request):
    AddressCollection(request.session).update()



import logging
log = logging.getLogger('onegov.winterthur')  # noqa
log.addHandler(logging.NullHandler())  # noqa

from translationstring import TranslationStringFactory  # noqa
_ = TranslationStringFactory('onegov.winterthur')  # noqa

from onegov.winterthur.app import WinterthurApp

__all__ = ('WinterthurApp', 'log', '_')



from onegov.winterthur.collections.address import AddressCollection
from onegov.winterthur.collections.address import AddressSubsetCollection

__all__ = ('AddressCollection', 'AddressSubsetCollection')



import attr
import textwrap

from .utils import as_choices
from .utils import build_metadata
from .utils import contains_department
from .utils import contains_email
from .utils import contains_location
from .utils import contains_only_numbers
from .utils import contains_street
from .utils import contains_website
from .utils import Geocoder
from .utils import html_to_text
from .utils import load_files_by_prefix
from .utils import pop_matching
from .utils import store_in_zip


def transform_vollzug(path, prefix, output_file):
    institutions = get_institutions(path, prefix)

    def unique_values(values, attr):
        unique = set()

        for value in values:
            unique.update(getattr(value, attr))

        return sorted(list(unique))

    topics = as_choices(unique_values(institutions, 'topics'))

    metadata = build_metadata(
        title='Vollzugsbehörden',
        lead='Die Vollzugsbehörden der Stadt Winterthur',
        title_format='[Fachstelle/Name] [Fachstelle/Abteilung]',
        lead_format='[Fachstelle/Department]',
        content_fields=[
            'Kategorie/Themen',
            'Kategorie/Vollzugsbereiche',
            'Kategorie/Vollzugsaufgaben'
        ],
        contact_fields=[
            'Fachstelle/Name',
            'Fachstelle/Abteilung',
            'Fachstelle/Department',
            'Kontakt/Adresse',
            'Kontakt/E-Mail',
            'Kontakt/Telefon',
            'Kontakt/Webseite'
        ],
        keyword_fields=[
            'Kategorie/Themen'
        ],
        structure=textwrap.dedent(f"""\
            # Fachstelle
            Name *= ___
            Abteilung = ___
            Department *= ___

            # Kontakt
            Adresse = ...
            Telefon = ___
            E-Mail = ___
            Webseite = ___

            # Kategorie
            Themen =
                {topics}

            Vollzugsbereiche = ...[16]
            Vollzugsaufgaben = ...[16]

        """))

    geo = Geocoder()

    data = [
        {
            'Fachstelle/Name': i.name,
            'Fachstelle/Department': i.department,
            'Fachstelle/Abteilung': i.branch,
            'Kontakt/Adresse': i.address,
            'Kontakt/E-Mail': i.email,
            'Kontakt/Telefon': i.phone,
            'Kontakt/Webseite': i.website,
            'Kategorie/Themen': list(i.topics),
            'Kategorie/Vollzugsbereiche': '\n'.join(
                f'- {c.strip()}' for c in sorted(list(i.chapters))
            ),
            'Kategorie/Vollzugsaufgaben': '\n'.join(
                f'- {t.strip()}' for t in sorted(list(i.tasks))
            ),
            'Latitude': geo.geocode(i.address).latitude,
            'Longitude': geo.geocode(i.address).longitude,
        } for i in institutions
    ]

    store_in_zip(output_file, {
        'metadata.json': metadata,
        'data.json': data
    })


def load_files(path, prefix):
    return load_files_by_prefix(path, prefix, mapping=(
        ('chapter_theme', (
            'chapter_uid',
            'chapter',
            'topic',
            'topic_pid'
        )),
        ('chapter_mm', (
            'uid_local',
            'uid_foreign',
            'sorting'
        )),
        ('institution_vollzugsbehoerden', (
            'uid',
            'institution',
            'description'
        ))
    ))


def get_institutions(path, prefix):
    chapters, links, institutions = load_files(path, prefix)

    # topics by chapter
    topics = {c.chapter: c.topic for c in chapters.lines}

    # chapters by institution id
    chapters = {c.chapter_uid: c.chapter for c in chapters.lines}
    chapters = {l.uid_local: chapters.get(l.uid_foreign) for l in links.lines}

    # resulting institutions
    result = {}

    for r in institutions.lines:
        task = r.institution

        for i in Institution.from_description(r.uid, r.description):
            i = result[i.title] = result.get(i.title, i)
            i.tasks.add(task)

            if chapters[r.uid]:
                i.chapters.add(chapters[r.uid])

            if chapters[r.uid] in topics:
                i.topics.add(topics[chapters[r.uid]])

    return sorted([i for i in result.values()], key=lambda i: i.title)


@attr.s(auto_attribs=True)
class Institution(object):
    name: str
    branch: str
    department: str
    address: str
    phone: str = None
    email: str = None
    website: str = None

    topics: set = attr.Factory(set)
    chapters: set = attr.Factory(set)
    tasks: set = attr.Factory(set)

    @property
    def title(self):
        return ' '.join(part for part in (
            self.name,
            self.branch
        ) if part)

    @classmethod
    def from_description(cls, uid, description):

        def clean(description):
            return description.replace('Tel.:', '')

        def split(description):
            for desc in description.split('</td'):
                if html_to_text(desc).strip():
                    yield desc

        for desc in split(clean(description)):

            lines = html_to_text(desc).split('\n')

            if len(lines) < 6:
                print(f"Could not parse a description for {uid}:")
                print(desc)
                continue

            yield cls(**match_address(lines))


def match_address(lines):
    department = pop_matching(lines, contains_department) or ''
    phone = pop_matching(lines, contains_only_numbers) or ''
    street = pop_matching(lines, contains_street) or ''
    location = pop_matching(lines, contains_location) or ''
    website = pop_matching(lines, contains_website) or ''
    email = pop_matching(lines, contains_email) or ''

    name = lines.pop(0)
    branch = lines and lines.pop(0) or ''

    return {
        'name': name,
        'branch': branch,
        'department': department,
        'address': '\n'.join((p for p in (street, location) if p)),
        'phone': phone,
        'website': website and f'http://{website}' or '',
        'email': email,
    }



from itertools import groupby
from onegov.core.cache import lru_cache
from onegov.core.security import Public, Private
from onegov.winterthur import WinterthurApp, _
from onegov.winterthur.collections import AddressCollection
from onegov.winterthur.collections import AddressSubsetCollection
from onegov.winterthur.layout import AddressLayout
from onegov.winterthur.layout import AddressSubsetLayout
from urllib.parse import quote_plus


@WinterthurApp.html(
    model=AddressCollection,
    permission=Public,
    template='streets.pt'
)
def view_streets(self, request):
    request.include('street-search')

    by_letter = {
        letter: tuple(streets) for letter, streets in groupby(
            self.streets(), lambda s: s.letter
        )
    }

    @lru_cache(maxsize=1)
    def link_to_street(street):
        return request.class_link(AddressSubsetCollection, {'street': street})

    return {
        'layout': AddressLayout(self, request),
        'title': _("Streets Directory"),
        'streets': by_letter,
        'link_to_street': link_to_street
    }


@WinterthurApp.view(
    model=AddressCollection,
    permission=Private,
    request_method='POST',
    name="update"
)
def update_streets(self, request):
    request.assert_valid_csrf_token()

    self.update()

    request.success(_("The streets directory has been updated"))


@WinterthurApp.html(
    model=AddressSubsetCollection,
    permission=Public,
    template='street.pt'
)
def view_street(self, request):
    def external_link_to_street(address):
        q = quote_plus(str(address.street_id))
        return f'https://stadtplan.winterthur.ch/?locate=strasse&locations={q}'

    def external_link_to_address(address):
        q = quote_plus(str(address.title))
        return f'https://stadtplan.winterthur.ch/?locate=adresse&locations={q}'

    return {
        'layout': AddressSubsetLayout(self, request),
        'title': self.street,
        'addresses': self.subset(),
        'parent': request.class_link(AddressCollection),
        'external_link_to_street': external_link_to_street,
        'external_link_to_address': external_link_to_address
    }
